"""Multi-point conversion functions between Kili and geojson formats."""


import uuid
from typing import Any, Dict, List, Optional


def geojson_multipoint_feature_to_kili_point_annotations(
    multipoint: Dict[str, Any],
    categories: Optional[List[Dict]] = None,
    children: Optional[Dict] = None,
) -> List[Dict[str, Any]]:
    """Convert a geojson multi-point feature to multiple Kili point annotations.

    Each point in the multi-point geometry is converted to a separate Kili marker annotation.
    All resulting annotations share the same categories and children but have unique mids.

    Args:
        multipoint: A geojson multi-point feature.
        categories: The categories of the annotations.
            If not provided, the categories are taken from the `kili` key of the geojson feature properties.
        children: The children of the annotations.
            If not provided, the children are taken from the `kili` key of the geojson feature properties.

    Returns:
        A list of Kili marker annotations, one for each point in the multi-point.

    !!! Example
        ```python
        >>> multipoint = {
            'type': 'Feature',
            'geometry': {
                'type': 'MultiPoint',
                'coordinates': [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]
            },
            'properties': {
                'kili': {
                    'categories': [{'name': 'landmark'}],
                    'children': {}
                }
            }
        }
        >>> geojson_multipoint_feature_to_kili_point_annotations(multipoint)
        [
            {
                'children': {},
                'categories': [{'name': 'landmark'}],
                'type': 'marker',
                'point': {'x': 1.0, 'y': 2.0},
                'mid': 'generated-uuid-1'
            },
            {
                'children': {},
                'categories': [{'name': 'landmark'}],
                'type': 'marker',
                'point': {'x': 3.0, 'y': 4.0},
                'mid': 'generated-uuid-2'
            },
            {
                'children': {},
                'categories': [{'name': 'landmark'}],
                'type': 'marker',
                'point': {'x': 5.0, 'y': 6.0},
                'mid': 'generated-uuid-3'
            }
        ]
        ```
    """

    assert (
        multipoint.get("type") == "Feature"
    ), f"Feature type must be `Feature`, got: {multipoint['type']}"
    assert (
        multipoint["geometry"]["type"] == "MultiPoint"
    ), f"Geometry type must be `MultiPoint`, got: {multipoint['geometry']['type']}"

    children = children or multipoint["properties"].get("kili", {}).get("children", {})
    categories = categories or multipoint["properties"]["kili"]["categories"]

    coords = multipoint["geometry"]["coordinates"]
    annotations = []

    for coord in coords:
        ret = {
            "children": children,
            "categories": categories,
            "type": "marker",
            "point": {"x": coord[0], "y": coord[1]},
            "mid": str(uuid.uuid4()),
        }
        annotations.append(ret)

    return annotations



from typing import Dict

from kili_formats.exceptions import NotCompatibleOptions
from kili_formats.media.common import scale_all_vertices


def scale_normalized_vertices_pdf_annotation(annotation: Dict, asset: Dict, **kwargs) -> None:
    """Scale normalized vertices of a PDF annotation.

    PDF annotations are different from image annotations because the asset width/height can vary.

    PDF only have BBox detection, so we only scale the boundingPoly and polys keys.
    """
    is_label_rotated = kwargs.get("is_label_rotated", False)

    if is_label_rotated:
        raise NotCompatibleOptions("PDF labels cannot be rotated")

    if "annotations" in annotation:
        # pdf annotations have two layers of "annotations"
        # https://docs.kili-technology.com/reference/export-object-entity-detection-and-relation#ner-in-pdfs
        for ann in annotation["annotations"]:
            scale_normalized_vertices_pdf_annotation(ann, asset, **kwargs)

    # an annotation has three keys:
    # - pageNumberArray: list of page numbers
    # - polys: list of polygons
    # - boundingPoly: list of bounding polygons
    # each polygon is a dict with a key "normalizedVertices" that is a list of vertices
    if "polys" in annotation and "boundingPoly" in annotation:
        try:
            page_number_to_dimensions = {
                page_resolution["pageNumber"]: {
                    "width": page_resolution["width"],
                    "height": page_resolution["height"],
                }
                for page_resolution in asset["pageResolutions"]
            }
        except (KeyError, TypeError) as err:
            raise NotCompatibleOptions(
                "PDF labels export with absolute coordinates require `pageResolutions` in the"
                " asset. Please use `kili.update_properties_in_assets(page_resolutions_array=...)`"
                " to update the page resolutions of your asset.`"
            ) from err

        for key in ("polys", "boundingPoly"):
            annotation[key] = [
                {
                    **value,  # keep the original normalizedVertices
                    "vertices": scale_all_vertices(
                        value["normalizedVertices"],
                        width=page_number_to_dimensions[page_number]["width"],
                        height=page_number_to_dimensions[page_number]["height"],
                    ),
                }
                for value, page_number in zip(annotation[key], annotation["pageNumberArray"])
            ]



"""Geometry collection conversion functions between Kili and geojson formats."""

import uuid
from typing import Any, Dict, List, Optional

from .line import geojson_linestring_feature_to_kili_line_annotation
from .multilinestring import geojson_multilinestring_feature_to_kili_line_annotations
from .multipoint import geojson_multipoint_feature_to_kili_point_annotations
from .point import geojson_point_feature_to_kili_point_annotation
from .polygon import geojson_polygon_feature_to_kili_polygon_annotation
from .segmentation import geojson_polygon_feature_to_kili_segmentation_annotation


def geojson_geometrycollection_feature_to_kili_annotations(
    geometrycollection: Dict[str, Any],
    categories: Optional[List[Dict]] = None,
    children: Optional[Dict] = None,
    mid: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """Convert a geojson geometry collection feature to Kili annotations.

    Processes each geometry in the collection and converts it to the appropriate Kili annotation type.
    Points become marker annotations, LineStrings become polyline annotations,
    Polygons become polygon or semantic annotations depending on the type specified,
    MultiPoint and MultiLineString geometries are expanded into multiple annotations.

    Args:
        geometrycollection: A geojson geometry collection feature.
        categories: The categories of the annotations.
            If not provided, the categories are taken from the `kili` key of the geojson feature properties.
        children: The children of the annotations.
            If not provided, the children are taken from the `kili` key of the geojson feature properties.
        mid: The mid of the annotations.
            If not provided, the mid is taken from the `id` key of the geojson feature.
            If no id is available, a new UUID is generated.

    Returns:
        A list of Kili annotations corresponding to each geometry in the collection.

    !!! Example
        ```python
        >>> geometrycollection = {
            'type': 'Feature',
            'geometry': {
                'type': 'GeometryCollection',
                'geometries': [
                    {
                        'type': 'Point',
                        'coordinates': [1.0, 2.0]
                    },
                    {
                        'type': 'LineString',
                        'coordinates': [[3.0, 4.0], [5.0, 6.0]]
                    },
                    {
                        'type': 'Polygon',
                        'coordinates': [[[7.0, 8.0], [9.0, 8.0], [9.0, 10.0], [7.0, 10.0], [7.0, 8.0]]]
                    }
                ]
            },
            'id': 'complex_001',
            'properties': {
                'kili': {
                    'categories': [{'name': 'complex'}],
                    'children': {}
                }
            }
        }
        >>> geojson_geometrycollection_feature_to_kili_annotations(geometrycollection)
        [
            {
                'children': {},
                'point': {'x': 1.0, 'y': 2.0},
                'categories': [{'name': 'complex'}],
                'mid': 'complex_001',
                'type': 'marker'
            },
            {
                'children': {},
                'polyline': [{'x': 3.0, 'y': 4.0}, {'x': 5.0, 'y': 6.0}],
                'categories': [{'name': 'complex'}],
                'mid': 'complex_001',
                'type': 'polyline'
            },
            {
                'children': {},
                'boundingPoly': [{'normalizedVertices': [{'x': 7.0, 'y': 8.0}, {'x': 9.0, 'y': 8.0}, {'x': 9.0, 'y': 10.0}, {'x': 7.0, 'y': 10.0}]}],
                'categories': [{'name': 'complex'}],
                'mid': 'complex_001',
                'type': 'polygon'
            }
        ]
        ```
    """

    assert (
        geometrycollection.get("type") == "Feature"
    ), f"Feature type must be `Feature`, got: {geometrycollection['type']}"
    assert (
        geometrycollection["geometry"]["type"] == "GeometryCollection"
    ), f"Geometry type must be `GeometryCollection`, got: {geometrycollection['geometry']['type']}"

    children = children or geometrycollection["properties"].get("kili", {}).get("children", {})
    categories = categories or geometrycollection["properties"]["kili"]["categories"]

    kili_properties = geometrycollection["properties"].get("kili", {})
    annotation_type = kili_properties.get("type")

    annotation_mid = None
    if mid is not None:
        annotation_mid = str(mid)
    elif "id" in geometrycollection:
        annotation_mid = str(geometrycollection["id"])
    else:
        annotation_mid = str(uuid.uuid4())

    geometries = geometrycollection["geometry"]["geometries"]
    annotations = []

    for geometry in geometries:
        feature = {"type": "Feature", "geometry": geometry, "properties": {"kili": kili_properties}}

        if geometry["type"] == "Point":
            if annotation_type and annotation_type != "marker":
                continue
            ann = geojson_point_feature_to_kili_point_annotation(
                feature, categories=categories, children=children, mid=annotation_mid
            )
            annotations.append(ann)

        elif geometry["type"] == "LineString":
            if annotation_type and annotation_type != "polyline":
                continue
            ann = geojson_linestring_feature_to_kili_line_annotation(
                feature, categories=categories, children=children, mid=annotation_mid
            )
            annotations.append(ann)

        elif geometry["type"] == "Polygon":
            if annotation_type:
                if annotation_type == "polygon":
                    ann = geojson_polygon_feature_to_kili_polygon_annotation(
                        feature, categories=categories, children=children, mid=annotation_mid
                    )
                    annotations.append(ann)
                elif annotation_type == "semantic":
                    anns = geojson_polygon_feature_to_kili_segmentation_annotation(
                        feature, categories=categories, children=children, mid=annotation_mid
                    )
                    annotations.extend(anns)
            else:
                ann = geojson_polygon_feature_to_kili_polygon_annotation(
                    feature, categories=categories, children=children, mid=annotation_mid
                )
                annotations.append(ann)

        elif geometry["type"] == "MultiPoint":
            if annotation_type and annotation_type != "marker":
                continue
            anns = geojson_multipoint_feature_to_kili_point_annotations(
                feature, categories=categories, children=children
            )
            annotations.extend(anns)

        elif geometry["type"] == "MultiLineString":
            if annotation_type and annotation_type != "polyline":
                continue
            anns = geojson_multilinestring_feature_to_kili_line_annotations(
                feature, categories=categories, children=children
            )
            annotations.extend(anns)

        elif geometry["type"] == "MultiPolygon":
            if annotation_type and annotation_type != "semantic":
                continue
            anns = geojson_polygon_feature_to_kili_segmentation_annotation(
                feature, categories=categories, children=children, mid=annotation_mid
            )
            annotations.extend(anns)

    return annotations



"""Polygon label utils."""

from typing import Any, Dict, List, Optional

from .exceptions import ConversionError


def get_oriented_area(vertices):
    """Returns the area value which gives an indication on the vertices order.
    Positive if counter-clockwise, negative if clockwise.

    This function uses the Shoelace formula :
    see also : https://en.wikipedia.org/wiki/Shoelace_formula
    """
    n = len(vertices)
    sum_product = 0

    for i in range(n):
        x1, y1 = vertices[i]["x"], vertices[i]["y"]
        x2, y2 = vertices[(i + 1) % n]["x"], vertices[(i + 1) % n]["y"]
        sum_product += (x1 - x2) * (y2 + y1)

    return sum_product


def order_counter_clockwise(vertices):
    """Returns the vertices, in the correct order :

    If the vertices are set clockwise, we reverse them to have them in the anti-clockwise order.
    For more information on the order expected for GeoJson :
    https://datatracker.ietf.org/doc/html/rfc7946#section-3.1.6
    """
    order = get_oriented_area(vertices)
    if order < 0:
        vertices.reverse()
    elif order == 0:
        raise ConversionError(
            f"Polygon order could not be identified as clockwise nor counter-clockwise because of \
                edges intersection in {vertices} and thus cannot be exported to GeoJson format."
        )

    return vertices


def kili_polygon_to_geojson_polygon(vertices: List[Dict[str, float]]) -> Dict[str, Any]:
    """Convert a Kili polygon to a geojson polygon.

    Args:
        vertices: Kili polygon vertices.

    Returns:
        A geojson polygon.

    !!! Example
        ```python
        >>> vertices = [
            {'x': 10.42, 'y': 27.12},
            {'x': 1.53, 'y': 14.57},
            {'x': 147.45, 'y': 14.12},
            {'x': 14.23, 'y': 0.23}
        ]
        >>> kili_polygon_to_geojson_polygon(vertices)
        {
            'type': 'Polygon',
            'coordinates': [
                [
                    [10.42, 27.12],
                    [1.53, 14.57],
                    [147.45, 14.12],
                    [14.23, 0.23],
                    [10.42, 27.12]
                ]
            ]
        }
        ```
    """
    reordered_polygon_vertices = order_counter_clockwise(vertices)
    polygon = [[vertex["x"], vertex["y"]] for vertex in reordered_polygon_vertices]

    polygon.append(polygon[0])  # the first and last positions must be the same
    return {"type": "Polygon", "coordinates": [polygon]}


def kili_polygon_annotation_to_geojson_polygon_feature(
    polygon_annotation: Dict[str, Any], job_name: Optional[str] = None
) -> Dict[str, Any]:
    """Convert a Kili polygon annotation to a geojson polygon feature.

    Args:
        polygon_annotation: a Kili polygon annotation.
        job_name: the name of the job to which the annotation belongs.

    Returns:
        A geojson polygon feature.

    !!! Example
        ```python
        >>> polygon = {
            'children': {},
            'boundingPoly': [{'normalizedVertices': [{'x': -79.0, 'y': -3.0}]}],
            'categories': [{'name': 'A'}],
            'mid': 'mid_object',
            'type': 'polygon'
        }
        >>> kili_polygon_annotation_to_geojson_polygon_feature(polygon, 'job_name')
        {
            'type': 'Feature',
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[-79.0, -3.0], [-79.0, -3.0]]]},
                'id': 'mid_object',
                'properties': {
                    'kili': {
                        'categories': [{'name': 'A'}],
                        'children': {},
                        'type': 'polygon',
                        'job': 'job_name'
                    }
                }
            }
        }
        ```
    """
    polygon = polygon_annotation
    assert (
        polygon["type"] == "polygon"
    ), f"Annotation type must be `polygon`, got: {polygon['type']}"

    ret = {
        "type": "Feature",
        "geometry": kili_polygon_to_geojson_polygon(
            polygon["boundingPoly"][0]["normalizedVertices"]
        ),
    }
    if "mid" in polygon:
        ret["id"] = polygon["mid"]
    ret["properties"] = {
        "kili": {k: v for k, v in polygon.items() if k not in ["boundingPoly", "mid"]}
    }
    if job_name is not None:
        ret["properties"]["kili"]["job"] = job_name
    return ret


def geojson_polygon_feature_to_kili_polygon_annotation(
    polygon: Dict[str, Any],
    categories: Optional[List[Dict]] = None,
    children: Optional[Dict] = None,
    mid: Optional[str] = None,
) -> Dict[str, Any]:
    """Convert a geojson polygon feature to a Kili polygon annotation.

    Args:
        polygon: a geojson polygon feature.
        categories: the categories of the annotation.
            If not provided, the categories are taken from the `kili` key of the geojson feature properties.
        children: the children of the annotation.
            If not provided, the children are taken from the `kili` key of the geojson feature properties.
        mid: the mid of the annotation.
            If not provided, the mid is taken from the `id` key of the geojson feature.


    Returns:
        A Kili polygon annotation.

    !!! Example
        ```python
        >>> polygon = {
            'type': 'Feature',
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[[-79.0, -3.0], [-79.0, -3.0]]]},
            },
            'id': 'mid_object',
            'properties': {
                'kili': {
                    'categories': [{'name': 'A'}],
                    'children': {},
                    'type': 'polygon',
                    'job': 'job_name'
                }
            }
        }
        >>> geojson_polygon_feature_to_kili_polygon_annotation(polygon)
        {
            'children': {},
            'boundingPoly': [{'normalizedVertices': [{'x': -79.0, 'y': -3.0}]}],
            'categories': [{'name': 'A'}],
            'mid': 'mid_object',
            'type': 'polygon'
        }
        ```
    """
    assert (
        polygon.get("type") == "Feature"
    ), f"Feature type must be `Feature`, got: {polygon['type']}"
    assert (
        polygon["geometry"]["type"] == "Polygon"
    ), f"Geometry type must be `Polygon`, got: {polygon['geometry']['type']}"

    children = children or polygon["properties"].get("kili", {}).get("children", {})
    categories = categories or polygon["properties"]["kili"]["categories"]

    ret = {
        "children": children,
        "categories": categories,
        "type": "polygon",
    }
    coords = polygon["geometry"]["coordinates"][0]
    normalized_vertices = [{"x": coord[0], "y": coord[1]} for coord in coords[:-1]]
    ret["boundingPoly"] = [{"normalizedVertices": normalized_vertices}]

    if mid is not None:
        ret["mid"] = str(mid)
    elif "id" in polygon:
        ret["mid"] = str(polygon["id"])

    return ret



"""Bounding box conversion functions between Kili and geojson formats."""

from typing import Any, Dict, List, Optional


def kili_bbox_to_geojson_polygon(vertices: List[Dict[str, float]]) -> Dict[str, Any]:
    """Convert a Kili bounding box to a geojson polygon.

    Args:
        vertices: Kili bounding polygon vertices.

    Returns:
        A geojson polygon.

    !!! Example
        ```python
        >>> vertices = [
            {'x': 12.0, 'y': 3.0},
            {'x': 12.0, 'y': 4.0},
            {'x': 13.0, 'y': 4.0},
            {'x': 13.0, 'y': 3.0}
        ]
        >>> kili_bbox_to_geojson_polygon(vertices)
        {
            'type': 'Polygon',
            'coordinates': [
                [
                    [12.0, 3.0],
                    [12.0, 4.0],
                    [13.0, 4.0],
                    [13.0, 3.0],
                    [12.0, 3.0]
                ]
            ]
        }
        ```
    """
    vertex_name_to_value = {}
    for vertex, point_name in zip(
        vertices, ("bottom_left", "top_left", "top_right", "bottom_right")
    ):
        vertex_name_to_value[point_name] = vertex

    ret = {"type": "Polygon", "coordinates": []}
    ret["coordinates"] = [
        [
            [vertex_name_to_value["bottom_left"]["x"], vertex_name_to_value["bottom_left"]["y"]],
            [vertex_name_to_value["bottom_right"]["x"], vertex_name_to_value["bottom_right"]["y"]],
            [vertex_name_to_value["top_right"]["x"], vertex_name_to_value["top_right"]["y"]],
            [vertex_name_to_value["top_left"]["x"], vertex_name_to_value["top_left"]["y"]],
            [vertex_name_to_value["bottom_left"]["x"], vertex_name_to_value["bottom_left"]["y"]],
        ]
    ]

    return ret


def kili_bbox_annotation_to_geojson_polygon_feature(
    bbox_annotation: Dict[str, Any], job_name: Optional[str] = None
) -> Dict[str, Any]:
    """Convert a Kili bounding box annotation to a geojson polygon feature.

    Args:
        bbox_annotation: a Kili bounding box annotation.
        job_name: the name of the job to which the annotation belongs.

    Returns:
        A geojson polygon feature.

    !!! Example
        ```python
        >>> bbox = {
            'children': {},
            'boundingPoly': [
                    {
                        'normalizedVertices': [
                            {'x': -12.6, 'y': 12.87},
                            {'x': -42.6, 'y': 22.17},
                            {'x': -17.6, 'y': -22.4},
                            {'x': 2.6, 'y': -1.87}
                        ]
                    }
                ],
            'categories': [{'name': 'A'}],
            'mid': 'mid_object',
            'type': 'rectangle'
        }
        >>> kili_bbox_annotation_to_geojson_polygon_feature(bbox, 'job_name')
        {
            'type': 'Feature',
            'geometry': {
                'type': 'Polygon',
                'coordinates': [
                    [
                        [-12.6, 12.87],
                        [-42.6, 22.17],
                        [-17.6, -22.4],
                        [2.6, -1.87],
                        [-12.6, 12.87]
                    ]
                ]
            },
            'id': 'mid_object',
            'properties': {
                'kili': {
                    'categories': [{'name': 'A'}],
                    'children': {},
                    'type': 'rectangle',
                    'job': 'job_name'
                }
            }
        }
        ```
    """
    bbox = bbox_annotation
    assert bbox["type"] == "rectangle", f"Annotation type must be `rectangle`, got: {bbox['type']}"

    ret = {
        "type": "Feature",
        "geometry": kili_bbox_to_geojson_polygon(bbox["boundingPoly"][0]["normalizedVertices"]),
    }
    if "mid" in bbox:
        ret["id"] = bbox["mid"]
    ret["properties"] = {
        "kili": {k: v for k, v in bbox.items() if k not in ["boundingPoly", "mid"]}
    }
    if job_name is not None:
        ret["properties"]["kili"]["job"] = job_name
    return ret


def geojson_polygon_feature_to_kili_bbox_annotation(
    polygon: Dict[str, Any],
    categories: Optional[List[Dict]] = None,
    children: Optional[Dict] = None,
    mid: Optional[str] = None,
) -> Dict[str, Any]:
    """Convert a geojson polygon feature to a Kili bounding box annotation.

    Args:
        polygon: a geojson polygon feature.
        categories: the categories of the annotation.
            If not provided, the categories are taken from the `kili` key of the geojson feature properties.
        children: the children of the annotation.
            If not provided, the children are taken from the `kili` key of the geojson feature properties.
        mid: the mid of the annotation.
            If not provided, the mid is taken from the `id` key of the geojson feature.

    Returns:
        A Kili bounding box annotation.

    !!! Example
        ```python
        >>> polygon = {
            'type': 'Feature',
            'geometry': {
                'type': 'Polygon',
                'coordinates': [
                    [
                        [-12.6, 12.87],
                        [-42.6, 22.17],
                        [-17.6, -22.4],
                        [2.6, -1.87],
                        [-12.6, 12.87]
                    ]
                ]
            },
            'id': 'mid_object',
            'properties': {
                'kili': {
                    'categories': [{'name': 'A'}],
                    'children': {},
                    'type': 'rectangle',
                    'job': 'job_name'
                }
            }
        }
        >>> geojson_polygon_feature_to_kili_bbox_annotation(polygon)
        {
            'children': {},
            'boundingPoly': [
                    {
                        'normalizedVertices': [
                            {'x': -12.6, 'y': 12.87},
                            {'x': -42.6, 'y': 22.17},
                            {'x': -17.6, 'y': -22.4},
                            {'x': 2.6, 'y': -1.87}
                        ]
                    }
                ],
            'categories': [{'name': 'A'}],
            'mid': 'mid_object',
            'type': 'rectangle'
        }
        ```
    """
    assert (
        polygon.get("type") == "Feature"
    ), f"Feature type must be `Feature`, got: {polygon['type']}"
    assert (
        polygon["geometry"]["type"] == "Polygon"
    ), f"Geometry type must be `Polygon`, got: {polygon['geometry']['type']}"

    children = children or polygon["properties"].get("kili", {}).get("children", {})
    categories = categories or polygon["properties"]["kili"]["categories"]

    ret = {
        "children": children,
        "categories": categories,
        "type": "rectangle",
    }
    # geojson polygon has one more point than kili bounding box
    coords = polygon["geometry"]["coordinates"][0]
    normalized_vertices = [
        {"x": coords[0][0], "y": coords[0][1]},
        {"x": coords[3][0], "y": coords[3][1]},
        {"x": coords[2][0], "y": coords[2][1]},
        {"x": coords[1][0], "y": coords[1][1]},
    ]
    ret["boundingPoly"] = [{"normalizedVertices": normalized_vertices}]

    if mid is not None:
        ret["mid"] = mid
    elif "id" in polygon:
        ret["mid"] = polygon["id"]

    return ret
